__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable

import numpy as np

from . import BaseExecutableDriver, FlatRecursiveMixin

if False:
    from ..types.sets import DocumentSet


class BaseIndexDriver(FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`add` by default """

    def __init__(self, executor: str = None, method: str = 'add', *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)

    def check_key_length(self, val: Iterable[str]):
        """
        Check if the max length of val(e.g. doc id) is larger than key_length.

        :param val: The values to be checked
        """
        m_val = max(len(v) for v in val)
        if m_val > self.exec.key_length:
            raise ValueError(f'{self.exec} allows only keys of length {self.exec.key_length}, '
                             f'but yours is {m_val}.')


class VectorIndexDriver(BaseIndexDriver):
    """Extracts embeddings and ids from the documents and forwards them to the executor.
    In case `method` is 'delete', the embeddings are ignored.
    If `method` is not 'delete', documents without content are filtered out.
    """

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        embed_vecs, docs_pts = docs.all_embeddings
        if docs_pts:
            keys = [doc.id for doc in docs_pts]
            self.check_key_length(keys)
            self.exec_fn(keys, np.stack(embed_vecs))


class KVIndexDriver(BaseIndexDriver):
    """Forwards pairs of serialized documents and ids to the executor."""

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        info = [(doc.id, doc.SerializeToString()) for doc in docs]
        if info:
            keys, values = zip(*info)
            self.check_key_length(keys)
            self.exec_fn(keys, values)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Any

from .running_stats import RunningStats
from .. import BaseExecutor
from ..compound import CompoundExecutor


class BaseEvaluator(BaseExecutor):
    """A :class:`BaseEvaluator` is used to evaluate different messages coming from any kind of executor
    """

    metric = ''  #: Get the name of the evaluation metric

    def post_init(self):
        """Initialize running stats."""
        super().post_init()
        self._running_stats = RunningStats()

    def evaluate(self, actual: Any, desired: Any, *args, **kwargs) -> float:
        """Evaluates difference between param:`actual` and `param:desired`, needs to be implemented in subclass."""
        raise NotImplementedError

    @property
    def mean(self) -> float:
        """Get the running mean."""
        return self._running_stats.mean

    @property
    def std(self) -> float:
        """Get the running standard variance."""
        return self._running_stats.std

    @property
    def variance(self) -> float:
        """Get the running variance."""
        return self._running_stats.variance


class FileBasedEvaluator(CompoundExecutor):

    """A Frequently used pattern for combining A :class:`BinaryPbIndexer` and :class:`BaseEvaluator`.
     It will be equipped with predefined ``requests.on`` behaviors:

         -  At evaluation time(query or index)
             - 1. Checks for the incoming document, gets its value from the `BinaryPbIndexer` and fills the `groundtruth of the request
             - 2. Filter the documents that do not have a corresponding groundtruth
             - 3. The BaseEvaluator works as if the `groundtruth` had been provided by the client as it comes in the request.

    .. warning::
        The documents that are not found to have an indexed groundtruth are removed from the `request` so that the `Evaluator` only
        works with documents which have groundtruth.

     One can use the :class:`FileBasedEvaluator` via

     .. highlight:: yaml
     .. code-block:: yaml

         !FileBasedEvaluator
         components:
           - !BinaryPbIndexer
             with:
               index_filename: ground_truth.gz
             metas:
               name: groundtruth_index  # a customized name
               workspace: ${{TEST_WORKDIR}}
           - !BaseEvaluator

     Without defining any ``requests.on`` logic. When load from this YAML, it will be auto equipped with

     .. highlight:: yaml
     .. code-block:: yaml

         on:
           [SearchRequest, IndexRequest]:
             - !LoadGroundTruthDriver
               with:
                 executor: BaseKVIndexer
             - !BaseEvaluateDriver
               with:
                 executor: BaseEvaluator
           ControlRequest:
             - !ControlReqDriver {}
     """



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BasePaddleEncoder


class ImagePaddlehubEncoder(BasePaddleEncoder):
    """
    :class:`ImagePaddlehubEncoder` encodes data from a ndarray,
    potentially B x (Channel x Height x Width) into a ndarray of `B x D`.

    Internally, :class:`ImagePaddlehubEncoder` wraps the models from `paddlehub`.
    https://github.com/PaddlePaddle/PaddleHub

    :param model_name: the name of the model. Supported models include
        ``xception71_imagenet``, ``xception65_imagenet``, ``xception41_imagenet``,
        ``vgg19_imagenet``, ``vgg16_imagenet``, ``vgg13_imagenet``, ``vgg11_imagenet``,
        ``shufflenet_v2_imagenet``,``se_resnext50_32x4d_imagenet``,
        ``se_resnext101_32x4d_imagenet``,  ``resnext50_vd_64x4d_imagenet``,
        ``resnext50_vd_32x4d_imagenet``, `resnext50_64x4d_imagenet``,
        ``resnext50_32x4d_imagenet``, ``resnext152_vd_64x4d_imagenet``,
        ``resnext152_64x4d_imagenet``, ``resnext152_32x4d_imagenet``,
        ``resnext101_vd_64x4d_imagenet``, ``resnext101_vd_32x4d_imagenet``,
        ``resnext101_32x8d_wsl``, ``resnext101_32x48d_wsl``, ``resnext101_32x32d_wsl``,
        ``resnext101_32x16d_wsl``, ``resnet_v2_50_imagenet``, ``resnet_v2_34_imagenet``,
        ``resnet_v2_18_imagenet``, ``resnet_v2_152_imagenet``, ``resnet_v2_101_imagenet``,
        ``mobilenet_v2_imagenet``, ``inception_v4_imagenet``, ``googlenet_imagenet``,
        ``efficientnetb7_imagenet``, ``efficientnetb6_imagenet``, ``efficientnetb5_imagenet``,
        ``efficientnetb4_imagenet``, ``efficientnetb3_imagenet``, ``efficientnetb2_imagenet``,
        ``efficientnetb1_imagenet``, ``efficientnetb0_imagenet``, ``dpn68_imagenet``,
        ``dpn131_imagenet``, ``dpn107_imagenet``, ``densenet264_imagenet``,
        ``densenet201_imagenet``, ``densenet169_imagenet``, ``densenet161_imagenet``,
        ``densenet121_imagenet``, ``darknet53_imagenet``, ``alexnet_imagenet``,
    :param pool_strategy: the pooling strategy. Default is `None`.
    :param channel_axis: The axis of the color channel, default is -3
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self,
                 model_name: str = None,
                 pool_strategy: str = None,
                 channel_axis: int = -3,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.pool_strategy = pool_strategy
        self.channel_axis = channel_axis
        self.model_name = model_name or 'xception71_imagenet'
        self.pool_strategy = pool_strategy or 'mean'
        self._default_channel_axis = -3
        self.inputs_name = None
        self.outputs_name = None

    def post_init(self):
        """Load model."""
        super().post_init()
        import paddlehub as hub
        module = hub.Module(name=self.model_name)
        inputs, outputs, self.model = module.context(trainable=False)
        self.get_inputs_and_outputs_name(inputs, outputs)
        self.exe = self.to_device()

    def close(self):
        self.exe.close()
        super().close()

    def get_inputs_and_outputs_name(self, input_dict, output_dict):
        """Get inputs_name (image name) and outputs_name (feature map)."""
        self.inputs_name = input_dict['image'].name
        self.outputs_name = output_dict['feature_map'].name
        if self.model_name.startswith('vgg') or self.model_name.startswith('alexnet'):
            self.outputs_name = f'@HUB_{self.model_name}@fc_1.tmp_2'

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode data into a ndarray of `B x D`. `
        B` is the batch size and `D` is the Dimension.

        :param data: a `B x T x (Channel x Height x Width)` numpy ``ndarray``,
            `B` is the size of the batch, `T` is the number of frames
        :return: a `B x D` numpy ``ndarray``, `D` is the output dimension
        """
        if self.channel_axis != self._default_channel_axis:
            data = np.moveaxis(data, self.channel_axis, self._default_channel_axis)
        feature_map, *_ = self.exe.run(
            program=self.model,
            fetch_list=[self.outputs_name],
            feed={self.inputs_name: data.astype('float32')},
            return_numpy=True
        )
        if feature_map.ndim == 2 or self.pool_strategy is None:
            return feature_map
        return self.get_pooling(feature_map)

    def get_pooling(self, data: 'np.ndarray') -> 'np.ndarray':
        """Get ndarray with selected pooling strategy"""
        _reduce_axis = tuple((i for i in range(len(data.shape)) if i > 1))
        return getattr(np, self.pool_strategy)(data, axis=_reduce_axis)



import numpy as np
from jina.executors.rankers import Chunk2DocRanker
from jina.helper import deprecated_alias

class SimpleAggregateRanker(Chunk2DocRanker):
    """
    :class:`SimpleAggregateRanker` aggregates the score
    of the matched doc from the matched chunks.
    For each matched doc, the score is aggregated
    from all the matched chunks belonging to that doc.

    :param: aggregate_function: defines the used aggregate function
        and can be one of: [min, max, mean, median, sum, prod]
    :param: inverse_score: plus-one inverse by 1/(1+score)
    :raises:
        ValueError: If `aggregate_function` is not any of the expected types
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    AGGREGATE_FUNCTIONS = ['min', 'max', 'mean', 'median', 'sum', 'prod']

    @deprecated_alias(is_reversed_score=('inverse_score', False))
    def __init__(self, aggregate_function: str, inverse_score: bool = False, *args, **kwargs):
        """Set constructor"""
        super().__init__(*args, **kwargs)
        self.inverse_score = inverse_score
        if aggregate_function in self.AGGREGATE_FUNCTIONS:
            self.np_aggregate_function = getattr(np, aggregate_function)
        else:
            raise ValueError(f'The aggregate function "{aggregate_function}" is not in "{self.AGGREGATE_FUNCTIONS}".')

    def _get_score(self, match_idx, query_chunk_meta, match_chunk_meta, *args, **kwargs):
        scores = match_idx[self.COL_SCORE]
        aggregated_score = self.np_aggregate_function(scores)
        if self.inverse_score:
            if aggregated_score < 0:
                raise ValueError(f'Setting "is_reversed_score" to True does not allow to have an aggregated document '
                                 f'negative score')
            aggregated_score = 1. / (1. + aggregated_score)
        return self.get_doc_id(match_idx), aggregated_score



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple, Iterable

from collections import defaultdict

import numpy as np

from . import ContextAwareRecursiveMixin, BaseRecursiveDriver, FlatRecursiveMixin
from ..types.sets import ChunkSet, MatchSet, DocumentSet


class ReduceAllDriver(ContextAwareRecursiveMixin, BaseRecursiveDriver):
    """:class:`ReduceAllDriver` merges chunks/matches from all requests, recursively.

    .. note::

        It uses the last request as a reference.
    """

    def __init__(self, traversal_paths: Tuple[str] = ('c',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_root(self, docs):
        request = self.msg.request
        request.body.ClearField('docs')
        request.docs.extend(docs)

    def _apply_all(self, doc_sequences: Iterable['DocumentSet'], *args, **kwargs) -> None:
        doc_pointers = {}
        for docs in doc_sequences:
            if isinstance(docs, (ChunkSet, MatchSet)):
                context_id = docs.reference_doc.id
                if context_id not in doc_pointers:
                    doc_pointers[context_id] = docs.reference_doc
                else:
                    if isinstance(docs, ChunkSet):
                        doc_pointers[context_id].chunks.extend(docs)
                    else:
                        doc_pointers[context_id].matches.extend(docs)
            else:
                self._apply_root(docs)


class CollectEvaluationDriver(FlatRecursiveMixin, BaseRecursiveDriver):
    """Merge all evaluations into one, grouped by ``doc.id`` """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(
            self,
            docs: 'DocumentSet',
            *args,
            **kwargs) -> None:
        doc_pointers = {}
        for doc in docs:
            if doc.id not in doc_pointers:
                doc_pointers[doc.id] = doc.evaluations
            else:
                doc_pointers[doc.id].extend(doc.evaluations)


class ConcatEmbedDriver(BaseRecursiveDriver):
    """Concat all embeddings into one, grouped by ``doc.id`` """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        """Performs the concatenation of all embeddings in `self.docs`.

        :param *args: *args not used. Only for complying with parent class interface.
        :param **kwargs: **kwargs not used. Only for complying with parent class interface.
        """
        all_documents = self.docs.traverse_flatten(self._traversal_paths)
        doc_pointers = self._collect_embeddings(all_documents)

        last_request_documents = self.req.docs.traverse_flatten(self._traversal_paths)
        self._concat_apply(last_request_documents, doc_pointers)

    def _collect_embeddings(self, docs: 'DocumentSet'):
        doc_pointers = defaultdict(list)
        for doc in docs:
            doc_pointers[doc.id].append(doc.embedding)
        return doc_pointers

    def _concat_apply(self, docs, doc_pointers):
        for doc in docs:
            doc.embedding = np.concatenate(doc_pointers[doc.id], axis=0)



from typing import Dict, List, Tuple
from collections import defaultdict, namedtuple

import numpy as np

from ....executors.rankers import Chunk2DocRanker, COL_STR_TYPE
from ....types.document import Document
from ....types.score import NamedScore

from .. import BaseRankDriver

if False:
    from ....types.sets import DocumentSet


class BaseAggregateMatchesRankerDriver(BaseRankDriver):
    """Drivers inherited from this Driver focus on aggregating scores from `chunks` to its `parents`."""

    def __init__(self,
                 keep_source_matches_as_chunks: bool = False,
                 *args,
                 **kwargs):
        """

        :param keep_source_matches_as_chunks: A flag to indicate if the driver must return the old matches of the query or its chunks
            (at a greater granularity level (k + 1)) as the chunks of the new computed `matches` (at granularity level k)
            Set it to `True` when keeping track of the chunks that lead to a retrieved result.
        :param *args: *args for super
        :param **kwargs: **kwargs for super

        .. note::
            When set `keep_source_matches_as_chunks=True`, the chunks of the match contains **ONLY** the chunks leading
            to the match rather than **ALL** the chunks of the match.

        """
        super().__init__(*args, **kwargs)
        self.keep_source_matches_as_chunks = keep_source_matches_as_chunks

    QueryMatchInfo = namedtuple('QueryMatchInfo', 'match_parent_id match_id query_id score')

    def _extract_query_match_info(self, match: Document, query: Document):
        return self.QueryMatchInfo(match_parent_id=match.parent_id,
                                   match_id=match.id,
                                   query_id=query.id,
                                   score=match.score.value)

    def _insert_query_matches(self,
                              query: Document,
                              parent_id_chunk_id_map: dict,
                              chunk_matches_by_id: dict,
                              docs_scores: 'np.ndarray'):
        """
        :param query: the query Document where the resulting matches will be inserted
        :param parent_id_chunk_id_map: a map with parent_id as key and list of previous matches ids as values
        :param chunk_matches_by_id: the previous matches of the query (at a higher granularity) grouped by the new map (by its parent)
        :param docs_scores: An `np.ndarray` resulting from the ranker executor with the `scores` of the new matches
        """

        op_name = self.exec.__class__.__name__
        for doc_id, score in docs_scores:
            m = Document(id=doc_id)
            m.score = NamedScore(op_name=op_name,
                                 value=score)
            if self.keep_source_matches_as_chunks:
                for match_chunk_id in parent_id_chunk_id_map[doc_id]:
                    m.chunks.append(chunk_matches_by_id[match_chunk_id])
            query.matches.append(m)


class Chunk2DocRankDriver(BaseAggregateMatchesRankerDriver):
    """Extract matches score from chunks and use the executor to compute the rank and assign the resulting matches to the
    level above.

    Input-Output ::
        Input:
        document: {granularity: k-1}
                |- chunks: {granularity: k}
                |    |- matches: {granularity: k}
                |
                |- chunks: {granularity: k}
                    |- matches: {granularity: k}
        Output:
        document: {granularity: k-1}
                |- chunks: {granularity: k}
                |    |- matches: {granularity: k}
                |
                |- chunks: {granularity: k}
                |    |- matches: {granularity: k}
                |
                |-matches: {granularity: k-1} (Ranked according to Ranker Executor)

        .. note::
            - It traverses on ``chunks`` not on ``matches``. This is because ranker needs context information
            from ``matches`` for several ``chunks``
    """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """
        :param docs: the doc which gets bubbled up matches
        :param *args: not used (kept to maintain interface)
        :param **kwargs: not used (kept to maintain interface)
        """
        for doc in docs:
            chunks = doc.chunks
            match_idx = []  # type: List[Tuple[str, str, str, float]]
            query_meta = {}  # type: Dict[str, Dict]
            match_meta = {}  # type: Dict[str, Dict]
            parent_id_chunk_id_map = defaultdict(list)
            matches_by_id = defaultdict(Document)
            for chunk in chunks:
                query_meta[chunk.id] = chunk.get_attrs(*self._exec_query_keys) if self._exec_query_keys else None
                for match in chunk.matches:
                    match_info = self._extract_query_match_info(match=match, query=chunk)
                    match_idx.append(match_info)
                    match_meta[match.id] = match.get_attrs(*self._exec_match_keys) if self._exec_match_keys else None
                    parent_id_chunk_id_map[match.parent_id].append(match.id)
                    matches_by_id[match.id] = match

            if match_idx:
                match_idx = np.array(match_idx,
                                     dtype=[
                                         (Chunk2DocRanker.COL_PARENT_ID, COL_STR_TYPE),
                                         (Chunk2DocRanker.COL_DOC_CHUNK_ID, COL_STR_TYPE),
                                         (Chunk2DocRanker.COL_QUERY_CHUNK_ID, COL_STR_TYPE),
                                         (Chunk2DocRanker.COL_SCORE, np.float64)
                                     ]
                                     )

                docs_scores = self.exec_fn(match_idx, query_meta, match_meta)

                self._insert_query_matches(query=doc,
                                           parent_id_chunk_id_map=parent_id_chunk_id_map,
                                           chunk_matches_by_id=matches_by_id,
                                           docs_scores=docs_scores)


class AggregateMatches2DocRankDriver(BaseAggregateMatchesRankerDriver):
    """This Driver is intended to take a `document` with matches at a `given granularity > 0`, clear those matches and substitute
    these matches by the documents at a lower granularity level.
    Input-Output ::
        Input:
        document: {granularity: k}
            |- matches: {granularity: k}

        Output:
        document: {granularity: k}
            |- matches: {granularity: k-1} (Sorted according to Ranker Executor)

    Imagine a case where we are querying a system with text documents chunked by sentences. When we query the system,
    we use sentences (chunks) to query it. So at some point we will have:
    `query sentence (documents of granularity 1):
        matches: indexed sentences (documents of level depth 1)`
    `
    But in the output we want to have the full document that better matches the `sentence`.
    `query sentence (documents of granularity 1):
        matches: indexed full documents (documents of granularity 0).
    `
    Using this Driver before querying a Binary Index with full binary document data can be very useful to implement a search system.
    """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """

        :param docs: the document at granularity ``k``
        :param *args: not used (kept to maintain interface)
        :param **kwargs: not used (kept to maintain interface)

        .. note::
            - This driver will substitute the ``matches`` of `docs` to the corresponding ``parent documents`` of its current ``matches`` according
            to the executor.
            - Set the ``traversal_paths`` of this driver to identify the documents, which needs to get bubbled up matches.
        """

        for doc in docs:
            matches = doc.matches

            match_idx = []
            query_meta = {}
            match_meta = {}
            parent_id_chunk_id_map = defaultdict(list)
            matches_by_id = defaultdict(Document)

            query_meta[doc.id] = doc.get_attrs(*self._exec_query_keys) if self._exec_query_keys else None

            for match in matches:
                match_info = self._extract_query_match_info(match=match, query=doc)
                match_idx.append(match_info)
                match_meta[match.id] = match.get_attrs(*self._exec_match_keys) if self._exec_match_keys else None
                parent_id_chunk_id_map[match.parent_id].append(match.id)
                matches_by_id[match.id] = match

            if match_idx:
                match_idx = np.array(match_idx,
                                     dtype=[
                                         (Chunk2DocRanker.COL_PARENT_ID, COL_STR_TYPE),
                                         (Chunk2DocRanker.COL_DOC_CHUNK_ID, COL_STR_TYPE),
                                         (Chunk2DocRanker.COL_QUERY_CHUNK_ID, COL_STR_TYPE),
                                         (Chunk2DocRanker.COL_SCORE, np.float64)
                                     ]
                                     )

                docs_scores = self.exec_fn(match_idx, query_meta, match_meta)
                # This ranker will change the current matches
                doc.ClearField('matches')
                self._insert_query_matches(query=doc,
                                           parent_id_chunk_id_map=parent_id_chunk_id_map,
                                           chunk_matches_by_id=matches_by_id,
                                           docs_scores=docs_scores)



"""Argparser module for remote runtime"""
from ...helper import add_arg_group
from .... import __default_host__
from ....enums import CompressAlgo
from .... import helper


def mixin_remote_parser(parser):
    """Add the options for remote expose
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='Expose')

    gp.add_argument('--host', type=str, default=__default_host__,
                    help=f'The host address of the runtime, by default it is {__default_host__}.')

    gp.add_argument('--port-expose',
                    type=int,
                    default=helper.random_port(),
                    help='The port of the host exposed to the public')


def mixin_grpc_parser(parser=None):
    """Add the options for gRPC
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='GRPC/REST')

    gp.add_argument('--max-message-size', type=int, default=-1,
                    help='The maximum send and receive size for gRPC server in bytes, -1 means unlimited')
    gp.add_argument('--proxy', action='store_true', default=False,
                    help='If set, respect the http_proxy and https_proxy environment variables. '
                         'otherwise, it will unset these proxy variables before start. '
                         'gRPC seems to prefer no proxy')
    gp.add_argument('--prefetch', type=int, default=50,
                    help='The number of pre-fetched requests from the client')
    gp.add_argument('--prefetch-on-recv', type=int, default=1,
                    help='The number of additional requests to fetch on every receive')
    gp.add_argument('--restful', '--rest-api', action='store_true', default=False,
                    help='If set, use RESTful interface instead of gRPC as the main interface')
    gp.add_argument('--compress', type=CompressAlgo.from_string, choices=list(CompressAlgo), default=CompressAlgo.LZ4,
                    help='''
The compress algorithm used over the entire Flow.

Note that this is not necessarily effective, it depends on the settings of `--compress-lwm` and `compress-hwm`''')
    gp.add_argument('--compress-min-bytes', type=int, default=1024,
                    help='The original message size must be larger than this number to trigger the compress algorithm, '
                         '-1 means disable compression.')
    gp.add_argument('--compress-min-ratio', type=float, default=1.1,
                    help='The compression ratio (uncompressed_size/compressed_size) must be higher than this number '
                         'to trigger the compress algorithm.')



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Models for Text and Image Composition."""

import numpy as np
import torch
import torchvision
import torch.nn.functional as F
import text_model, torch_functions


class ConCatModule(torch.nn.Module):

  def __init__(self):
    super(ConCatModule, self).__init__()

  def forward(self, x):
    x = torch.cat(x, dim=1)
    return x


class ImgTextCompositionBase(torch.nn.Module):
  """Base class for image + text composition."""

  def __init__(self):
    super(ImgTextCompositionBase, self).__init__()
    self.normalization_layer = torch_functions.NormalizationLayer(
        normalize_scale=4.0, learn_scale=True)
    self.soft_triplet_loss = torch_functions.TripletLoss()

  def extract_img_feature(self, imgs):
    raise NotImplementedError

  def extract_text_feature(self, texts):
    raise NotImplementedError

  def compose_img_text(self, imgs, texts):
    raise NotImplementedError

  def compute_loss(self,
                   imgs_query,
                   modification_texts,
                   imgs_target,
                   soft_triplet_loss=True):
    mod_img1 = self.compose_img_text(imgs_query, modification_texts)
    mod_img1 = self.normalization_layer(mod_img1)
    img2 = self.extract_img_feature(imgs_target)
    img2 = self.normalization_layer(img2)
    assert (mod_img1.shape[0] == img2.shape[0] and
            mod_img1.shape[1] == img2.shape[1])
    if soft_triplet_loss:
      return self.compute_soft_triplet_loss_(mod_img1, img2)
    else:
      return self.compute_batch_based_classification_loss_(mod_img1, img2)

  def compute_soft_triplet_loss_(self, mod_img1, img2):
    triplets = []
    labels = range(mod_img1.shape[0]) + range(img2.shape[0])
    for i in range(len(labels)):
      triplets_i = []
      for j in range(len(labels)):
        if labels[i] == labels[j] and i != j:
          for k in range(len(labels)):
            if labels[i] != labels[k]:
              triplets_i.append([i, j, k])
      np.random.shuffle(triplets_i)
      triplets += triplets_i[:3]
    assert (triplets and len(triplets) < 2000)
    return self.soft_triplet_loss(torch.cat([mod_img1, img2]), triplets)

  def compute_batch_based_classification_loss_(self, mod_img1, img2):
    x = torch.mm(mod_img1, img2.transpose(0, 1))
    labels = torch.tensor(range(x.shape[0])).long()
    labels = torch.autograd.Variable(labels).cuda()
    return F.cross_entropy(x, labels)


class ImgEncoderTextEncoderBase(ImgTextCompositionBase):
  """Base class for image and text encoder."""

  def __init__(self, texts, embed_dim):
    super(ImgEncoderTextEncoderBase, self).__init__()

    # img model
    img_model = torchvision.models.resnet18(pretrained=True)

    class GlobalAvgPool2d(torch.nn.Module):

      def forward(self, x):
        return F.adaptive_avg_pool2d(x, (1, 1))

    img_model.avgpool = GlobalAvgPool2d()
    img_model.fc = torch.nn.Sequential(torch.nn.Linear(512, embed_dim))
    self.img_model = img_model

    # text model
    self.text_model = text_model.TextLSTMModel(
        texts_to_build_vocab=texts,
        word_embed_dim=embed_dim,
        lstm_hidden_dim=embed_dim)

  def extract_img_feature(self, imgs):
    return self.img_model(imgs)

  def extract_text_feature(self, texts):
    return self.text_model(texts)


class SimpleModelImageOnly(ImgEncoderTextEncoderBase):

  def compose_img_text(self, imgs, texts):
    return self.extract_img_feature(imgs)


class SimpleModelTextOnly(ImgEncoderTextEncoderBase):

  def compose_img_text(self, imgs, texts):
    return self.extract_text_feature(texts)


class Concat(ImgEncoderTextEncoderBase):
  """Concatenation model."""

  def __init__(self, texts, embed_dim):
    super(Concat, self).__init__(texts, embed_dim)

    # composer
    class Composer(torch.nn.Module):
      """Inner composer class."""

      def __init__(self):
        super(Composer, self).__init__()
        self.m = torch.nn.Sequential(
            torch.nn.BatchNorm1d(2 * embed_dim), torch.nn.ReLU(),
            torch.nn.Linear(2 * embed_dim, 2 * embed_dim),
            torch.nn.BatchNorm1d(2 * embed_dim), torch.nn.ReLU(),
            torch.nn.Dropout(0.1), torch.nn.Linear(2 * embed_dim, embed_dim))

      def forward(self, x):
        f = torch.cat(x, dim=1)
        f = self.m(f)
        return f

    self.composer = Composer()

  def compose_img_text(self, imgs, texts):
    img_features = self.extract_img_feature(imgs)
    text_features = self.extract_text_feature(texts)
    return self.compose_img_text_features(img_features, text_features)

  def compose_img_text_features(self, img_features, text_features):
    return self.composer((img_features, text_features))


class TIRG(ImgEncoderTextEncoderBase):
  """The TIGR model.

  The method is described in
  Nam Vo, Lu Jiang, Chen Sun, Kevin Murphy, Li-Jia Li, Li Fei-Fei, James Hays.
  "Composing Text and Image for Image Retrieval - An Empirical Odyssey"
  CVPR 2019. arXiv:1812.07119
  """

  def __init__(self, texts, embed_dim):
    super(TIRG, self).__init__(texts, embed_dim)

    self.a = torch.nn.Parameter(torch.tensor([1.0, 10.0, 1.0, 1.0]))
    self.gated_feature_composer = torch.nn.Sequential(
        ConCatModule(), torch.nn.BatchNorm1d(2 * embed_dim), torch.nn.ReLU(),
        torch.nn.Linear(2 * embed_dim, embed_dim))
    self.res_info_composer = torch.nn.Sequential(
        ConCatModule(), torch.nn.BatchNorm1d(2 * embed_dim), torch.nn.ReLU(),
        torch.nn.Linear(2 * embed_dim, 2 * embed_dim), torch.nn.ReLU(),
        torch.nn.Linear(2 * embed_dim, embed_dim))

  def compose_img_text(self, imgs, texts):
    img_features = self.extract_img_feature(imgs)
    text_features = self.extract_text_feature(texts)
    return self.compose_img_text_features(img_features, text_features)

  def compose_img_text_features(self, img_features, text_features):
    f1 = self.gated_feature_composer((img_features, text_features))
    f2 = self.res_info_composer((img_features, text_features))
    f = torch.sigmoid(f1) * img_features * self.a[0] + f2 * self.a[1]
    return f


class TIRGLastConv(ImgEncoderTextEncoderBase):
  """The TIGR model with spatial modification over the last conv layer.

  The method is described in
  Nam Vo, Lu Jiang, Chen Sun, Kevin Murphy, Li-Jia Li, Li Fei-Fei, James Hays.
  "Composing Text and Image for Image Retrieval - An Empirical Odyssey"
  CVPR 2019. arXiv:1812.07119
  """

  def __init__(self, texts, embed_dim):
    super(TIRGLastConv, self).__init__(texts, embed_dim)

    self.a = torch.nn.Parameter(torch.tensor([1.0, 10.0, 1.0, 1.0]))
    self.mod2d = torch.nn.Sequential(
        torch.nn.BatchNorm2d(512 + embed_dim),
        torch.nn.Conv2d(512 + embed_dim, 512 + embed_dim, [3, 3], padding=1),
        torch.nn.ReLU(),
        torch.nn.Conv2d(512 + embed_dim, 512, [3, 3], padding=1),
    )

    self.mod2d_gate = torch.nn.Sequential(
        torch.nn.BatchNorm2d(512 + embed_dim),
        torch.nn.Conv2d(512 + embed_dim, 512 + embed_dim, [3, 3], padding=1),
        torch.nn.ReLU(),
        torch.nn.Conv2d(512 + embed_dim, 512, [3, 3], padding=1),
    )

  def compose_img_text(self, imgs, texts):
    text_features = self.extract_text_feature(texts)

    x = imgs
    x = self.img_model.conv1(x)
    x = self.img_model.bn1(x)
    x = self.img_model.relu(x)
    x = self.img_model.maxpool(x)

    x = self.img_model.layer1(x)
    x = self.img_model.layer2(x)
    x = self.img_model.layer3(x)
    x = self.img_model.layer4(x)

    # mod
    y = text_features
    y = y.reshape((y.shape[0], y.shape[1], 1, 1)).repeat(
        1, 1, x.shape[2], x.shape[3])
    z = torch.cat((x, y), dim=1)
    t = self.mod2d(z)
    tgate = self.mod2d_gate(z)
    x = self.a[0] * torch.sigmoid(tgate) * x + self.a[1] * t

    x = self.img_model.avgpool(x)
    x = x.view(x.size(0), -1)
    x = self.img_model.fc(x)
    return x



from typing import Tuple, Optional

import numpy as np

from .. import BaseExecutableDriver, FlatRecursiveMixin
from ...types.document import Document
from ...types.score import NamedScore

if False:
    from ...types.sets import DocumentSet


class BaseRankDriver(FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`rank` by default """

    def __init__(self, executor: Optional[str] = None, method: str = 'score', *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)

    @property
    def _exec_match_keys(self):
        """ Property to provide backward compatibility to executors relying in `required_keys`
        :return: keys for attribute lookup in matches
        """
        return self.exec.match_required_keys if hasattr(self.exec, 'match_required_keys') else getattr(self.exec,
                                                                                                       'required_keys',
                                                                                                       None)

    @property
    def _exec_query_keys(self):
        """ Property to provide backward compatibility to executors relying in `required_keys`

        :return: keys for attribute lookup in matches
        """
        return self.exec.query_required_keys if hasattr(self.exec, 'query_required_keys') else getattr(self.exec,
                                                                                                       'required_keys',
                                                                                                       None)


class Matches2DocRankDriver(BaseRankDriver):
    """ This driver is intended to only resort the given matches on the 0 level granularity for a document.
    It gets the scores from a Ranking Executor, which does only change the scores of matches.
    Afterwards, the Matches2DocRankDriver resorts all matches for a document.
    Input-Output ::
        Input:
        document: {granularity: 0, adjacency: k}
            |- matches: {granularity: 0, adjacency: k+1}
        Output:
        document: {granularity: 0, adjacency: k}
            |- matches: {granularity: 0, adjacency: k+1} (Sorted according to scores from Ranker Executor)
    """

    def __init__(self, reverse: bool = False, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)
        self.reverse = reverse

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """

        :param docs: the matches of the ``context_doc``, they are at granularity ``k``
        :param *args: not used (kept to maintain interface)
        :param **kwargs: not used (kept to maintain interface)

        .. note::
            - This driver will change in place the ordering of ``matches`` of the ``context_doc`.
            - Set the ``traversal_paths`` of this driver such that it traverses along the ``matches`` of the ``chunks`` at the level desired.
        """
        for doc in docs:
            query_meta = doc.get_attrs(*self._exec_query_keys) if self._exec_query_keys else None

            matches = doc.matches
            old_match_scores = {match.id: match.score.value for match in matches}
            match_meta = {
                match.id: match.get_attrs(*self._exec_match_keys) for match in matches
            } if self._exec_match_keys else None

            # if there are no matches, no need to sort them
            if not old_match_scores:
                continue

            new_match_scores = self.exec_fn(query_meta, old_match_scores, match_meta)
            self._sort_matches_in_place(doc, new_match_scores)

    def _sort_matches_in_place(self, context_doc: 'Document', match_scores: 'np.ndarray') -> None:
        op_name = self.exec.__class__.__name__
        cm = context_doc.matches
        cm.build()
        for match_id, score in match_scores:
            cm[match_id].score = NamedScore(value=score, op_name=op_name, ref_id=context_doc.id)

        cm.sort(key=lambda x: x.score.value, reverse=True)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

schema_requests = {
    'Jina::Requests': {
        'description': 'Define how the Executor behaves under network requests.',
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'use_default': {
                'description': 'If set, then inherit from the default `Driver` settings for this type of Executor.',
                'type': 'boolean',
                'default': False
            },
            'on': {
                'description': 'Defines how the `Executor` behaves under different types of request',
                'type': 'object',
                'properties': {
                    f'{r_type}Request': {
                        'type': 'object',
                        'properties': {
                            'with': {
                                'type': 'object',
                                'description': 'The common kwargs that all drivers defined under this Request.'
                            },
                            'drivers': {
                                '$ref': f'#/definitions/Jina::Drivers::All'
                            }
                        },
                        'additionalProperties': False,
                        'description': f'Defines how the `Executor` behaves under {r_type} request.'
                    } for r_type in ['Index', 'Train', 'Search', 'Update', 'Delete', 'Control']
                },
                'additionalProperties': False
            },
        }
    }
}



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable, Tuple

from ...types.querylang.queryset.dunderkey import dunder_get
from .. import QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver

if False:
    from ...types.sets import DocumentSet


class SortQL(QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver):
    """Sorts the incoming of the documents by the value of a given field.
     It can also work in reverse mode

        Example::
        - !ReduceAllDriver
            with:
                traversal_paths: ['m']
        - !SortQL
            with:
                reverse: true
                field: 'score__value'
                traversal_paths: ['m']
        - !SliceQL
            with:
                start: 0
                end: 50
                traversal_paths: ['m']

        `SortQL` will ensure that only the documents are sorted by the score value before slicing the first top 50 documents
    """

    def __init__(self, field: str, reverse: bool = False, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        """
        :param field: the value of the field drives the sort of the iterable docs
        :param reverse: sort the value from big to small
        :param traversal_paths: the traversal paths
        :param *args: *args
        :param **kwargs: **kwargs
        """

        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)
        self._reverse = reverse
        self._field = field

    def _apply_all(self, doc_sequences: Iterable['DocumentSet'], *args, **kwargs) -> None:
        for docs in doc_sequences:
            docs.sort(key=lambda x: dunder_get(x, self.field), reverse=self.reverse)



from typing import Callable, Dict, Any, Optional, List, Union

from pydantic import Field, BaseModel, create_model

from jina.enums import DataInputType
from jina.types.document import Document
from jina.parsers import set_client_cli_parser
from jina.proto.jina_pb2 import DocumentProto, QueryLangProto


class JinaStatusModel(BaseModel):
    """Pydantic BaseModel for Jina status, used as the response model in REST app."""
    jina: Dict
    envs: Dict
    used_memory: str


def build_model_from_pb(name: str, pb_model: Callable):
    """
    Build model from protobuf message.

    :param name: Name of the model.
    :param pb_model: protobuf message.
    :return: Model.
    """
    from google.protobuf.json_format import MessageToDict

    dp = MessageToDict(pb_model(), including_default_value_fields=True)

    all_fields = {k: (name if k in ('chunks', 'matches') else type(v), Field(default=v)) for k, v in dp.items()}
    if pb_model == QueryLangProto:
        all_fields['parameters'] = (Dict, Field(default={}))

    return create_model(name, **all_fields)


JinaDocumentModel = build_model_from_pb('Document', DocumentProto)
JinaDocumentModel.update_forward_refs()
JinaQueryLangModel = build_model_from_pb('QueryLang', QueryLangProto)
default_request_size = set_client_cli_parser().parse_args([]).request_size


class JinaRequestModel(BaseModel):
    """
    Jina request model.

    The base model for Jina REST request.
    """
    # To avoid an error while loading the request model schema on swagger, we've added an example.
    data: Union[List[JinaDocumentModel], List[Dict[str, Any]], List[str], List[bytes]] = \
        Field(..., example=[Document().dict()])
    request_size: Optional[int] = default_request_size
    mime_type: Optional[str] = ''
    queryset: Optional[List[JinaQueryLangModel]] = None
    data_type: DataInputType = DataInputType.AUTO


class JinaIndexRequestModel(JinaRequestModel):
    """Index request model."""
    pass


class JinaSearchRequestModel(JinaRequestModel):
    """Search request model."""
    pass


class JinaUpdateRequestModel(JinaRequestModel):
    """Update request model."""
    pass


class JinaDeleteRequestModel(JinaRequestModel):
    """Delete request model."""
    data: List[str]


class JinaControlRequestModel(JinaRequestModel):
    """Control request model."""
    pass


class JinaTrainRequestModel(JinaRequestModel):
    """Train request model."""
    pass
